from isogroup.base.feature import Feature
from isocor.base import LabelledChemical
from isogroup.base.misc import Misc
import pandas as pd


class Database:

    def __init__(self, dataset: pd.DataFrame, tracer="13C"):
        self.dataset = dataset
        self.features: list = []
        self.tracer: str = tracer
        self._tracer_element, self._tracer_idx = Misc._parse_strtracer(tracer)
        self.clusters: list = []

        _isodata: dict = LabelledChemical.DEFAULT_ISODATA
        self._delta_mz_tracer: float = _isodata[self._tracer_element]["mass"][1] - _isodata[
            self._tracer_element]["mass"][0]
        self._delta_mz_hydrogen: float = _isodata["H"]["mass"][0]

        self.initialize_theoretical_features()
        self.export_database(filename="Database_isotopic.tsv")

    def __len__(self) -> int:
        return len(self.dataset)

    def initialize_theoretical_features(self):

        """
        Creates chemical labelled from isocor functions
        then initializes the theoretical features from a database file
        """
        for _, line in self.dataset.iterrows():
            chemical = LabelledChemical(
                formula=line["formula"],
                tracer=self.tracer,
                derivative_formula="",
                tracer_purity=[1.0, 0.0],
                correct_NA_tracer=False,
                data_isotopes=None,
                charge=line["charge"],
                label=line["metabolite"],
            )
            for i in range(chemical.formula[self._tracer_element] + 1):
                mz = (chemical.molecular_weight + i * self._delta_mz_tracer
                      + line["charge"] * self._delta_mz_hydrogen)
                feature = Feature(
                    rt=line["rt"],
                    mz=mz,
                    tracer=self.tracer,
                    intensity=None,
                    chemical=[chemical],
                    isotopologue=[i],
                    metabolite=[chemical.label],
                    formula = line["formula"],
                )
                self.features.append(feature)


    def export_database(self, filename = None):
        """
        Create a DataFrame to summarize the database
        Optionnal: Export the DataFrame to a tsv file if a filename is provided with samples in column
        """

        # Create a DataFrame to summarize the theoretical features
        feature_data = []
        for feature in self.features:
            feature_data.append({
                "mz": feature.mz,
                "rt": feature.rt,
                "metabolite": ', '.join(feature.metabolite),
                "isotopologue": ', '.join(map(str, feature.isotopologue)),
                "formula": feature.formula,
                })

        df = pd.DataFrame(feature_data)

        # Export the DataFrame to a tsv file if a filename is provided
        if filename:
            df.to_csv(filename, sep="\t", index=False)

            return df


from isogroup.base.misc import Misc

class Feature:

    def __init__(self, rt: float, mz: float, tracer: str, intensity:float|None, feature_id = None, counter_formula: list|None=None, formula: list|None=None, sample: str|None=None,
                 chemical: list|None=None, metabolite: list|None=None, isotopologue: list|None=None, mz_error: list|None=None, rt_error: list|None=None, **extra_dims):
        self.rt = float(rt)
        self.mz = float(mz)
        self.tracer = tracer
        self._tracer_element, self._tracer_idx = Misc._parse_strtracer(tracer) 
        self.intensity = intensity
        self.feature_id = feature_id
        self.chemical = chemical if chemical is not None else []
        self.counter_formula = [i.formula for i in self.chemical] if self.chemical is not None else formula #formula ou [] ?
        self.formula = formula if formula is not None else []
        self.sample = sample
        self.mz_error = mz_error if mz_error is not None else []
        self.rt_error = rt_error if rt_error is not None else []
        self.metabolite = [i.label for i in self.chemical] if self.chemical is not None else metabolite #metabolite ou [] ?
        self.isotopologue = isotopologue if isotopologue is not None else []
        self.__dict__.update(extra_dims)
        self.is_adduct: tuple[bool, str] = (False, "")
        self.in_cluster = []


    def __repr__(self) -> str:
        """
        Return a string representation of the feature.
        :return: str
        """
        return (f"Feature(ID = {self.feature_id}, RT={self.rt}, Metabolite={self.metabolite}, Isotopologue={self.isotopologue}, "
                f"mz={self.mz}, "
                f"intensity={self.intensity})")
    
    # @property
    # def in_cluster(self):
    #     """
    #     Check if the feature is in another cluster
    #     Return the cluster_id if the feature is in it
    #     """
    #     pass





import re
from isocor.base import LabelledChemical

class Misc:

    @staticmethod
    def _parse_strtracer(str_tracer):
        """Parse the tracer code.

        Args:
            str_tracer (str): tracer code (e.g. "13C")

        Returns:
            tuple
                - (str) tracer element (e.g. "C")
                - (int) tracer index in :py:attr:`~data_isotopes`
        """
        try:
            tracer = re.search(r'(\d*)([A-Z][a-z]*)', str_tracer)
            count = int(tracer.group(1))
            tracer_el = tracer.group(2)
        except (ValueError, AttributeError):
            raise ValueError("Invalid tracer code: '{}'."
                             " Please check your inputs.".format(str_tracer))
        best_diff = float("inf")
        idx_tracer = None
        unexpected_msg = "Unexpected tracer code. Are you sure this isotope is "\
                         "in data_isotopes? '{}'".format(str_tracer)
        assert tracer_el in LabelledChemical.DEFAULT_ISODATA, unexpected_msg
        for i, mass in enumerate(LabelledChemical.DEFAULT_ISODATA[tracer_el]["mass"]):
            test_diff = abs(mass - count)
            if test_diff < best_diff:
                best_diff = test_diff
                idx_tracer = i
        assert best_diff < 0.5, unexpected_msg
        return (tracer_el, idx_tracer)


import pandas as pd
from isogroup.base.database import Database
from isogroup.base.feature import Feature
from isogroup.base.cluster import Cluster
from isogroup.base.misc import Misc


class Experiment:

    def __init__(self, dataset: pd.DataFrame, database: 'Database' = None, tracer=None):
        self.dataset = dataset
        self.database = database
        self.samples: dict = {} 
        self._mz_tol: None | float = None
        self._rt_tol: None | float = None

        self._tracer: None | str = tracer
        self._tracer_element, self._tracer_idx = Misc._parse_strtracer(tracer)
        self.clusters = None
        

    @property
    def rt_tol(self):
        return self._rt_tol

    @property
    def tracer(self):
        return self._tracer

    @property
    def tracer_element(self):
        return self._tracer_element

    @property
    def mz_tol(self):
        return self._mz_tol
    

    def initialize_experimental_features(self):
        """
        Initialize the experimental features from the dataset
        """
        for idx, _ in self.dataset.iterrows():
            mz = idx[0]
            rt = idx[1]
            id = idx[2]

            # Extract the intensity for each sample in the dataset
            for sample in self.dataset.columns:
                if sample not in ["mz", "rt", "id"]:
                    intensity = self.dataset.loc[idx, sample]

                    # Initialize the experimental features for each sample
                    feature = Feature(
                        rt=rt, mz=mz, tracer=self.tracer,
                        feature_id=id, 
                        intensity=intensity,
                        sample=sample
                        )
                    
                    # Add the feature in the list corresponding to the sample
                    if sample not in self.samples:
                        self.samples[sample] = {}
                    self.samples[sample][id] = feature


    def annotate_features(self, mz_tol, rt_tol):
        """
        Annotate the experiment features with the database within a given tolerance
        Calculate the mz error and the rt error
        """

        for sample in self.samples.values():
            for feature in sample.values():
                    
                for db_feature in self.database.features:

                    # Calculate the exact mz and rt errors
                    mz_error = (db_feature.mz - feature.mz)
                    rt_error = (db_feature.rt - feature.rt)

                    # Covert mz_error to ppm 
                    mz_error = (mz_error / feature.mz) * 1e6

                    # Check if the experimental feature is within tolerance
                    if abs(mz_error) <= mz_tol and abs(rt_error) <= rt_tol:
                        feature.chemical.append(db_feature.chemical[0])
                        feature.isotopologue.append(db_feature.isotopologue[0])
                        feature.metabolite.append(db_feature.chemical[0].label)
                        feature.formula.append(db_feature.chemical[0].formula)
                        feature.mz_error.append(mz_error)
                        feature.rt_error.append(rt_error)

        self._mz_tol = mz_tol
        self._rt_tol = rt_tol


    def annotate_experiment(self, mz_tol, rt_tol):
        """
        Annotate the experiment features with the database within a given tolerance
        MultiIndex DataFrame
        """
        # Initialize the experimental features from the dataset
        self.initialize_experimental_features()

        # Annotate the experimental features
        self.annotate_features(mz_tol, rt_tol)



    def export_features(self, filename = None, sample_name = None):
        """
        Create a DataFrame to summarize the annotated data
        Optionnal: Export the DataFrame to a tsv file if a filename is provided with samples in column
        Optionnal: Export the Dataframe of only one sample if a sample name is provided
        """

        # Create a DataFrame to summarize the experimental features
        feature_data = []
        for sample in self.samples.values():
            for feature in sample.values():
                feature_data.append({
                    "feature_id": feature.feature_id,
                    "mz": feature.mz,
                    "rt": feature.rt,
                    "metabolite": feature.metabolite,
                    "isotopologue": feature.isotopologue,
                    "mz_error": feature.mz_error,
                    "rt_error": feature.rt_error,
                    "sample": feature.sample,
                    "intensity": feature.intensity
                })

        # Create a DataFrame to summarize the annotated data
        df = pd.DataFrame(feature_data)

        # Export the DataFrame to a tsv file if a filename is provided
        if filename:
            df.to_csv(filename, sep="\t", index=False)

            # Export the Dataframe of only one sample if a sample name is provided
            if sample_name:
                df = df[df["sample"] == sample_name] # Filter the DataFrame by sample name
                df.to_csv(filename, sep="\t", index=False)

        return df

    def clusterize(self):
        """
        Create unique clusters from annotated features based on their names.
        """
        cluster_names = []
        
        # Group features by metabolite
        for sample in self.samples.values():
            for feature in sample.values():
                cluster_names += feature.metabolite

        cluster_names = set(cluster_names)

        # Create unique clusters
        self.clusters = {}

        for sample in self.samples.keys():
            self.clusters[sample] = {}
            for i,c in enumerate(cluster_names):
                features = self.get_features_from_name(c, sample)

                # Sort features by isotopologues
                features.sort(key=lambda f: f.isotopologue)

                # Assign the cluster_id to the features in the cluster
                for feature in features:
                    if not hasattr(feature, "in_cluster") or feature.in_cluster is None:
                        feature.in_cluster = [] 
                    feature.in_cluster.append(f"C{i}")  

                self.clusters[sample][c] = Cluster(features=features, cluster_id=f"C{i}", name=c)


    def get_features_from_name(self, name, sample_name:str):
        """
        Get a feature from the experiment by its name, in a given sample if provided
        """
        features = []
        for feature in self.samples[sample_name].values():
            if name in feature.metabolite:
                features.append(feature)
        return features

    
    def export_clusters(self, filename = None, sample_name = None):
        """
        Create a DataFrame to summarize the annotated clusters
        Optionnal: Export the DataFrame to a tsv file if a filename is provided
        Optionnal: Export the Dataframe of only one sample if a sample name is provided
        """
        
        # Check if the sample name is in the DataFrame
        all_samples = list(self.samples.keys())
        if sample_name is not None:
            if sample_name not in all_samples:
                raise ValueError(f"Sample {sample_name} not found in annotated clusters. Available samples: {', '.join(all_samples)}")
        
        cluster_data = []
        for sample, clusters in self.clusters.items():
            if sample_name is None or sample_name == sample: # Filter the DataFrame by sample name if provided
                for cname, cluster in clusters.items():
                    for feature in cluster.features:
                        idx = [i for i,j in enumerate(feature.metabolite) if j == cname][0]
                        # Get the cluster_id of the features in another cluster
                        other_clusters = [c.cluster_id for cluster_name, c in clusters.items() if feature in c.features and c.cluster_id != cluster.cluster_id]
                        cluster_data.append({
                            "cluster_id": cluster.cluster_id,
                            "metabolite": cluster.name,
                            "feature_id": feature.feature_id,
                            "mz": feature.mz,
                            "rt": feature.rt,
                            "feature_potential_metabolite": feature.metabolite,
                            "isotopologue": feature.isotopologue[idx],
                            "mz_error": feature.mz_error[idx],
                            "rt_error": feature.rt_error[idx],
                            "sample": feature.sample,
                            "intensity": feature.intensity,
                            "status": cluster.status,
                            "missing_isotopologue": cluster.missing_isotopologues,
                            "duplicated_isotopologue": cluster.duplicated_isotopologues,
                            # "in_cluster": feature.in_cluster,
                            "in_another_cluster": other_clusters
                        })

        # Create a DataFrame to summarize the annotated clusters
        df = pd.DataFrame(cluster_data)

        # Export the DataFrame to a tsv file if a filename is provided
        if filename:
            df.to_csv(filename, sep="\t", index=False)

        return df


    def get_clusters_from_name(self, name, sample_name:str):
        """
        Get a cluster from the experiment by its name, in a given sample if provided
        """
        for cluster in self.clusters[sample_name].values():
            if cluster.name == name:
                return cluster
        return None
    

    def clusters_summary(self, filename = None):
        """
        Export a tsv file with a summary of the clusters
        """
        # List to store the cluster summary data
        cluster_summary = []
        cluster_id_unique = set() # To store unique cluster_id

        for sample, clusters in self.clusters.items():
            for cluster in clusters.values():

                # Check if the cluster_id is unique
                if cluster.cluster_id not in cluster_id_unique:
                    cluster_id_unique.add(cluster.cluster_id)

                    summary = cluster.cluster_summary

                    # Retrieve the samples in which the cluster is present
                    samples_in_cluster = {sample for sample, clusters in self.clusters.items() if cluster.cluster_id in [c.cluster_summary["cluster_id"] for c in clusters.values()]}
                    summary["samples"] = len(samples_in_cluster)

                    cluster_summary.append(summary)

        # Create a DataFrame with the collected information
        df = pd.DataFrame(cluster_summary)

        # Export the DataFrame to a tsv file if a filename is provided
        if filename:
            df.to_csv(filename, sep="\t", index=False)

        return df



import argparse
from isogroup.base.database import Database
from isogroup.base.experiment import Experiment
from pathlib import Path
import pandas as pd


def process(args):

    # load data file
    inputdata = Path(args.inputdata)

    if not inputdata.exists():
        msg = f"File {inputdata} does not exist"
        raise FileNotFoundError(msg)

    # load database file
    if hasattr(args, 'D'):
        database = Path(args.D)
    else:
        msg = "No database file provided"
        raise ValueError(msg)
    if not database.exists():
        msg = f"File {database} does not exist"
        raise FileNotFoundError(msg)

    mztol = float(getattr(args, 'mztol', None))
    rttol = float(getattr(args, 'rttol', None))
    tracer = getattr(args, 'tracer', None)

    db_data = pd.read_csv(database, sep=";")

    database = Database(dataset=db_data, tracer="13C")
    data = pd.read_csv(inputdata, sep="\t")
    data = data.set_index(["mz", "rt", "id"])

    experiment = Experiment(dataset=data, database=database, tracer=tracer)
    experiment.annotate_experiment(mz_tol=mztol, rt_tol=rttol)
    experiment.clusterize()

    # Set working directory from output path
    if args.output:
        output = Path(args.output)
        experiment.export_features(filename=output.with_suffix('.features.tsv'))
        experiment.export_clusters(filename=output.with_suffix('.clusters.tsv'))
        experiment.clusters_summary(filename=output.with_suffix('.clusters_summary.tsv'))
    else:
        msg = "No output file provided"
        raise ValueError(msg)

def parseArgs():
    parser = argparse.ArgumentParser(argument_default=argparse.SUPPRESS,
                                     description='annotation of isotopic datasets')

    parser.add_argument("inputdata", help="measurements file to process")
    parser.add_argument("-D", type=str, help="path to database file (csv)") 

    parser.add_argument("-t", "--tracer", type=str, required=True,
                        help='the isotopic tracer (e.g. "13C")')
    parser.add_argument("--mztol", type=float, required=True,
                        help='mz tolerance in ppm (e.g. "5")')
    parser.add_argument("--rttol", type=float, required=True,
                        help='rt tolerance (e.g. "10")')
    parser.add_argument("-o", "--output", type=str,
                        help='output file for the clusters')
    # parser.add_argument("-v", "--verbose",
    #                     help="flag to enable verbose logs", action='store_true')
    return parser


def start_cli():
    parser = parseArgs()
    args = parser.parse_args()
    process(args)


from typing import List, Union, Iterator, Self
from isogroup.base.feature import Feature

class Cluster:

    def __init__(self, features: list|None, cluster_id = None, name=None):
        self.features = features
        self.cluster_id = cluster_id
        self.tracer_element = features[0]._tracer_element if features is not None else None
        self.tracer = features[0].tracer if features is not None else None
        self.name = name
        self._formula = None

    def __repr__(self) -> str:
        return f"Cluster({self.cluster_id}, {self.features})"
    
    def __len__(self) -> int:
        """
        Returns the number of features in the cluster
        """
        return len(self.features)


    def __add__(self, other: Union[Feature, Self]) -> Self:
        """
        Extend the cluster with either an extra feature or another cluster.
        :param other: Feature or Cluster
        :return: self
        """
        pass

    def __iadd__(self, other: Union[Feature, Self]) -> Self:
        """
        Extend the cluster with either an extra feature or another cluster.
        :param other: Feature or Cluster
        :return: self
        """
        pass

    def __remove__(self, other: Union[Feature, Self]) -> Self:
        """
        Remove a feature from the cluster
        :param other: Feature
        :return: self
        """
        pass

    def __contains__(self, item):
        pass

    def __iter__(self) -> Iterator[Feature]:
        return iter(self.features)
    
    
    @property
    def lowest_rt(self) -> float:
        return min([f.rt for f in self.features])

    @property
    def highest_rt(self) -> float:
        return max([f.rt for f in self.features])

    @property
    def lowest_mz(self) -> float:
        return min([f.mz for f in self.features])

    @property
    def highest_mz(self) -> float:
        return max([f.mz for f in self.features])

    @property
    def metabolite(self):
        """
        Returns the complete annotation for each feature in the cluster
        """
        return [f.metabolite for f in self.features]
    
    @property
    def chemical(self):
        """
        Returns the chemical object of the cluster
        """
        for feature in self.features:
            if self.name in feature.metabolite:
                idx = feature.metabolite.index(self.name)
                return feature.chemical[idx]

    @property
    def element_number(self):
        """
        Returns the number of tracer element for the formula matching to the cluster name
        """
        self.formula
        return self._formula[self.tracer_element]

    
    @property
    def isotopologues(self) -> List[int]:
        """
        Returns the list of isotopologues in the cluster
        """
        isotopologues = []
        for feature in self.features:
            if self.name in feature.metabolite:
                idx = feature.metabolite.index(self.name)
                isotopologues.append(feature.isotopologue[idx])
        return isotopologues


    @property
    def formula(self) -> str:
        """
        Returns the formula of the metabolite matching to the cluster name
        """
        if self._formula is None:
            # Check if the cluster is annotated
            if self.name is None:
                raise ValueError("No cluster found. Run clusterize() first")
            
            if self.name in self.features[0].metabolite:
                idx = self.features[0].metabolite.index(self.name)
                self._formula = self.features[0].formula[idx]
            else:  
                raise ValueError(f"No metabolite matching '{self.name}' found in the cluster features")
            
        return self._formula


    @property
    def expected_isotopologues_in_cluster(self):
        """
        Returns the list of expected isotopologues in the cluster
        Based on the number of tracer element in its formula
        """
        return list(range(self.element_number + 1))
                           

    @property
    def is_complete(self) -> bool:
        """
        Returns True if the cluster is complete (i.e contains all isotopologues expected)
        """   
        return len(self) == self.element_number + 1 and self.expected_isotopologues_in_cluster == self.isotopologues
    
    @property
    def is_incomplete(self) -> bool:
        """
        Returns True if the cluster is incomplete (i.e contains not all isotopologues expected)
        """
        return len(self) < self.element_number + 1 or len(set(self.isotopologues)) != len(self.expected_isotopologues_in_cluster)

    @property
    def is_duplicated(self) -> bool:
        """
        Returns True if the cluster contains duplicated isotopologues
        """
        return len(set(self.isotopologues)) != len(self.isotopologues)

    @property
    def is_corrupted(self) -> bool:
        """
        Returns True if the cluster is corrupted (overfilled ?) (i.e contains isotopologues not expected)
        """ 
        pass


    @property 
    def status(self):
        """
        Returns the cluster status (Ok, incomplete, duplicated, corrupted)
        """
        if self.is_complete:
            return "Ok"
        
        if self.is_incomplete and self.is_duplicated:
            return "Incomplete, duplicated isotopologues"

        if self.is_incomplete:
            return "Incomplete"

        if self.is_duplicated:
            return "Duplicated isotopologues"
        
        # if self.is_corrupted:
        #     return "Corrupted"
        

    @property
    def missing_isotopologues(self) -> List[int]:
        """
        Returns a list of missing isotopologues in the cluster
        """
        if self.is_incomplete:
            return [i for i in self.expected_isotopologues_in_cluster if i not in self.isotopologues]
        

    @property
    def duplicated_isotopologues(self):
        """
        Returns a list of duplicated isotopologues in the cluster
        """
        if self.is_duplicated:
            return [i for i in set(self.isotopologues) if self.isotopologues.count(i) > 1]
        
    
    @property
    def is_adduct(self) -> tuple[bool, str]:
        pass

    @property
    def cluster_summary(self):
        """
        Returns a summary of the cluster
        """
        return {
            "cluster_id": self.cluster_id,
            "name": self.name,
            "number_of_features": len(self),
            "isotopologues": self.isotopologues,
            "status": self.status,
            "missing_isotopologues": self.missing_isotopologues,
            "duplicated_isotopologues": self.duplicated_isotopologues
        }


import pandas as pd
from isogroup.base.feature import Feature


class Sample:
    def __init__(self, dataset: pd.DataFrame, sample_name : str, sample_type = None):
        self.data = dataset 
        self.sample_name = sample_name # Name of the sample
        self.sample_type = sample_type # Type of the sample : control, TP, fully labelled, etc.
        self.features = [] # List of features in the sample

    def initialize_features(self, annotated_data):
        """
        Initialize the features for this sample from the annotated data.
        For each feature, create a new instance with the intensity of the sample.
        """
        pass

    def __repr__(self):
        """
        Return a string representation of the sample.
        :return: str
        """
        return f"Sample({self.sample_name}, {self.sample_type}, {self.features})"
    